#-----------------------------------------------------------------------------
# Bugs-n-Beans  (descended from Swarm)
# 1/31/05
# Copyright (c) 2006 David G. Drumright (ockham)
# Fixed up some 11/11/06
#-----------------------------------------------------------------------------
'''
Jan 21,05
new thought about swarm, filling space, making 'poof'.
If the container is built properly, we can fill it
nicely by the following:

Find the center of the container.
For the particle, find its center and outer radius at initial scale.
Put each copy of the particle into the center of the container.
Randomly pick a vertex of the container.
Send the copy out toward that vertex, to a random radius along
the center-to-vert vector.

Provide a set of 'good' containers to use for such
things as a leaf pile.  For the leaf-pile poof,
have a morph or pose that expands the container to
a realistic up-and-out position, then another
morph or pose that lays it back down in with a
wider xz.
For an open-top container, provide a temporary
container that will shape the top of the pile
to be flat, with a morph to raise the angle of fall.
You can then remove or hide this container
to leave the top of the pile open.

The number and size of particles can be set at
startup; default or min count is the vert count
of the container, and default size is container volume / vert count.

1/26 working OK.
Maybe would be nice to have a move-only mode.  Would work
OK for the parented version.  If you don't hit Select,
we just take all the particles currently parented to
the selected Container, and move them again.

'''
#-----------------------------------------------------------------------------

from Tkinter import *
import tkFileDialog
import poser
import random
import string
import math
import os
random.seed()
random.seed()
scene = poser.Scene()
BasePathUncut=poser.AppLocation()
BasePath=BasePathUncut[:(-len('poser.exe'))]
PropPath=os.path.normpath(BasePath+'/Runtime/Libraries/props/*.PP2')

#-----------------------------------------------------------------------------
# Next section is TK
#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        master.title("Bugs -n- Beans")
        self.PartNameList=[] # Needed to hold info between select and go
        self.SamplePart=None # So we can detect no file

        self.StatusEntry = Entry(self.master,width=40)
        self.StatusEntry.grid(row=0,column=0)
        self.SelectPartButton = Button(self.master, text="Select particle", command=self.HandleSelect)
        self.SelectPartButton.grid(row=1, column=0)

        self.SliderFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.SliderFrame.grid(row=3,column=0,sticky=W+E)
        self.ActionFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ActionFrame.grid(row=4,column=0,sticky=W+E)
        self.ButtonFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ButtonFrame.grid(row=5,column=0,sticky=W+E)

        # Go buttons.
        self.GoButton = Button(self.ButtonFrame, text="Go", command=self.HandleGo)
        self.GoButton.grid(row=0, column=0)
        self.NoButton = Button(self.ButtonFrame, text="Cancel", command=self.HandleCancel)
        self.NoButton.grid(row=0, column=2)

        Label(self.SliderFrame, text="Number:").grid(row=0,column=0,sticky=E)
        self.CountSlider=Scale(self.SliderFrame, from_ = 1, to = 100,orient=HORIZONTAL)
        # Note that this starts and defaults at 1.
        self.CountSlider.grid(row=0, column=1)

        Label(self.SliderFrame, text="Size, percent of base prop:").grid( row=1,column=0,sticky=E)
        self.WantedSizeSlider=Scale(self.SliderFrame, from_ = 0.1, to = 100,orient=HORIZONTAL,resolution=0.1)
        self.WantedSizeSlider.grid(row=1,column=1)
        self.WantedSizeSlider.set(10) # Default of 10

        Label(self.SliderFrame, text="Random size:").grid( row=2,column=0,sticky=E)
        self.RandSizeSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandSizeSlider.grid(row=2,column=1)

        Label(self.SliderFrame, text="Random orientation:").grid( row=3,column=0,sticky=E)
        self.RandDirSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandDirSlider.grid(row=3,column=1)

        Label(self.ActionFrame, text="Random motion:").grid( row=0,column=0,sticky=E)
        self.SpeedSlider=Scale(self.ActionFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.SpeedSlider.grid(row=0,column=1)

        self.StartLabel=Label(self.ActionFrame, text="Start frame:")
        self.StartLabel.grid( row=1,column=0,sticky=E)
        self.StartEntry=Entry(self.ActionFrame, width=6)
        self.StartEntry.grid(row=1,column=1)

        self.EndLabel=Label(self.ActionFrame, text="End frame:")
        self.EndLabel.grid( row=2,column=0,sticky=E)
        self.EndEntry=Entry(self.ActionFrame, width=6)
        self.EndEntry.grid(row=2,column=1)

        # Set the default values for frame entries
        self.StartEntry.insert(0,str(1))
        self.EndEntry.insert(0,str(scene.NumFrames()))

        # Set up the Base first.
        # The base must be the selected item.
        self.BaseItem = scene.CurrentActor()
        # If it's part of a figure, get the figure as well.
        self.BaseFig = self.BaseItem.ItsFigure()
        if (not self.BaseItem.IsBodyPart()) and (not self.BaseItem.IsProp()):
            self.ShowStatus("Container prop not a good type.  Cancel and restart.")
        else:
            self.ShowStatus("First select the 'particle' prop.")

        # Get vert count of container to serve as default particlecount.
        try:
            self.BaseGeom=self.BaseItem.Geometry()
            Vcount=self.BaseGeom.NumVertices()
        except:
            self.ShowStatus("Container appears to be a BODY. Cancel and restart.")
            return

        # Set up close action
        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleSelect(self): # Get the filename only.
        # Pop up file box to get a Particle.
        self.FileName=tkFileDialog.askopenfilename(initialfile=PropPath, filetypes=[("Props", "PP2")] )
        if not self.FileName:
            self.ShowStatus("No file selected. Try again.")
            return

        # The actual value settings of the sample particle will be done along with the others
        # in HandleGo.
        self.ShowStatus("Now set values then hit Go.")


    # - - - - - - - - - - - - - - - - - -

    def SetupAll(self):
        if not self.FileName:
            self.ShowStatus("Select a particle first!")
            return
        else:
            self.ShowStatus("Starting setup...")

        scene.SetFrame(0)

        self.PartNameList=[] # Holds just the names.  The only reason we need this is
        # because we load one example first to get orig size, then load the
        # others in a separate loop.

        # Fetch one copy of the particle, so we can get its
        # bounding box in orig state.  This copy will become the
        # first instance; the rest (if Count>1) will just take
        # the data.
        try:
            scene.LoadLibraryProp(os.path.normpath(self.FileName))
        except:
            self.ShowStatus("Prop not usable. Try again.")
            return

        # The new prop will be CurrentActor at this moment.
        self.SamplePart = scene.CurrentActor()
        self.PartNameList.append( (self.SamplePart.Name(),0) )

        # Get the sliders we need here.
        Count = self.CountSlider.get()
        WantedSize = float(self.WantedSizeSlider.get()) / 100.0
        RandSize = float(self.RandSizeSlider.get()) / 1000.0
        RandDir  = float(self.RandDirSlider.get())  / 100.0
        Speed = float(self.SpeedSlider.get()) / 100.0

        scene.DrawAll() # Always needed when using WVs

        # Get the verts of the Container.
        WVcount=self.BaseGeom.NumVertices()
        WVs=[]
        MaxX=-1000.0
        MaxY=-1000.0
        MaxZ=-1000.0
        MinX= 1000.0
        MinY= 1000.0
        MinZ= 1000.0

        CX=0.0
        CY=0.0
        CZ=0.0
        for i in range(WVcount):
            WV=self.BaseGeom.WorldVertex(i)
            X=WV.X()
            if   X > MaxX: MaxX = X
            elif X < MinX: MinX = X
            CX=CX+X
            Y=WV.Y()
            if   Y > MaxY: MaxY = Y
            elif Y < MinY: MinY = Y
            CY=CY+Y
            Z=WV.Z()
            if   Z > MaxZ: MaxZ = Z
            elif Z < MinZ: MinZ = Z
            CZ=CZ+Z
            WVs.append( (i,X,Y,Z) )
        # Locate the center
        CX=CX/float(WVcount)
        CY=CY/float(WVcount)
        CZ=CZ/float(WVcount)

        # BaseRadius is max in each direction.
        BRx=abs(MaxX-MinX)/2.0
        BRy=abs(MaxY-MinY)/2.0
        BRz=abs(MaxZ-MinZ)/2.0

        # Now get the bounding box of the part in its loaded state.
        MaxX=-10000.0
        MaxY=-10000.0
        MaxZ=-10000.0
        MinX= 10000.0
        MinY= 10000.0
        MinZ= 10000.0
        scene.DrawAll()
        PartGeom=self.SamplePart.Geometry()
        for i in range(PartGeom.NumVertices()):
            WV=PartGeom.WorldVertex(i)
            X=WV.X()
            Y=WV.Y()
            Z=WV.Z()
            if   X < MinX: MinX = X
            elif X > MaxX: MaxX = X
            if   Y < MinY: MinY = Y
            elif Y > MaxY: MaxY = Y
            if   Z < MinZ: MinZ = Z
            elif Z > MaxZ: MaxZ = Z


        # Get the orig scale values.
        OrigAscale = self.SamplePart.ParameterByCode(poser.kParmCodeASCALE).Value()
        #  Figure the default scale on the basis of Desired Percent, Orig Scale,
        # and Original Percent.
        # Default setting of Ascale should = Orig Ascale * Desired Percent / Orig Percent,
        # where Orig Percent = PartRadius / Container RAdius.  Try pyth:
        # In this case we don't need the midpoint
        PRx  = (MaxX - MinX) / 2.0
        PRy  = (MaxY - MinY) / 2.0
        PRz  = (MaxZ - MinZ) / 2.0
        PartRadius = math.sqrt(PRx*PRx+PRy*PRy+PRz*PRz)
        BaseRadius = math.sqrt(BRx*BRx+BRy*BRy+BRz*BRz)
        OrigPercent = PartRadius / BaseRadius
        self.DefaultScale = OrigAscale * WantedSize / OrigPercent


        # Create the rest of the particles:
        for i in range(1,Count+1): # Start from 1 because we already have the first
            # Create a copy of the Part.
            try:
                scene.LoadLibraryProp(os.path.normpath(self.FileName))
            except:
                self.ShowStatus("Prop not usable. Try again.")
                return
            self.ShowStatus("Loading particle %d" % i)
            # The new prop will be currentactor at this moment.
            Part = scene.CurrentActor()
            # Pick one WV index at random and assign it to the particle.
            (i, vX,vY,vZ) = random.choice(WVs)
            self.PartNameList.append( (Part.Name(), i) )


        for (PartName,WVindex) in self.PartNameList: # List includes the sample
            Part = scene.Actor(PartName)
            self.ShowStatus("Locating and scaling %s" % PartName)
            # For convenience, get the necessary parms of all piecees:
            PartXrotParm =   Part.ParameterByCode(poser.kParmCodeXROT)
            PartYrotParm =   Part.ParameterByCode(poser.kParmCodeYROT)
            PartZrotParm =   Part.ParameterByCode(poser.kParmCodeZROT)

            PartXtranParm =  Part.ParameterByCode(poser.kParmCodeXTRAN)
            PartYtranParm =  Part.ParameterByCode(poser.kParmCodeYTRAN)
            PartZtranParm =  Part.ParameterByCode(poser.kParmCodeZTRAN)

            PartAscaleParm = Part.ParameterByCode(poser.kParmCodeASCALE)

            # Set the Y and X rot with desired degree of randomness. (No point
            # in setting all three; gimbal lock and allstufflikethatthere.)
            PartYrotParm.SetValue(PartYrotParm.Value()+360.0 * RandDir * random.random())
            PartXrotParm.SetValue(PartXrotParm.Value()+360.0 * RandDir * random.random())

            # Set the scale with desired degree of randomness.
            R=self.DefaultScale + RandSize * (random.random()-0.5)
            if R < 0.001: R = 0.001
            PartAscaleParm.SetValue(R)

            # Now place the locations, using vectors from center of
            # container to each vert of container as 'sliders'

            # Pick a proportion at random, pushing it outward just a bit.
            Pro = random.random()+0.4
            if Pro > 1.0: Pro = 1.0

            (i,WVx,WVy,WVz)=WVs[WVindex] # Reget the proper WV

            # Locate the particle at this spot
            PartXtranParm.SetValue(WVx*Pro)
            PartYtranParm.SetValue(WVy*Pro)
            PartZtranParm.SetValue(WVz*Pro)

    # - - - - - - - - - - - - - - - - - -

    def Swarm(self):
        Speed = float(self.SpeedSlider.get()) / 500.0
        RandSize = float(self.RandSizeSlider.get()) / 1000.0
        RandDir  = float(self.RandDirSlider.get())  / 100.0

        # Fetch start and end:
        StartFrame=1
        EndFrame=scene.NumFrames()
        s = self.StartEntry.get()
        if s:
            StartFrame = int(s)
        s = self.EndEntry.get()
        if s:
            EndFrame = int(s)
        if (EndFrame - StartFrame) < 2: return

        # Keyframe the particles outside the range of action to
        # avoid overshoots
        for (PartName,WVindex) in self.PartNameList:
            Part = scene.Actor(PartName)
            for F in range(0,StartFrame+1):
                Part.AddKeyFrame(F)
            for F in range(EndFrame,scene.NumFrames()):
                Part.AddKeyFrame(F)
        # Hitting every frame with random makes way too much wiggle.
        # Try hitting every 1/4 second or so, slightly random.
        Step = scene.FramesPerSecond()/4

        for F in range(StartFrame,EndFrame+1,Step):
            # Set the frame ahead or behind the step randomly:
            scene.SetFrame(F + int(random.uniform(-1.1,1.1))-1 )
            scene.DrawAll() # Always needed when using WVs
            self.ShowStatus("Working frame %d" % F)

            # Get the verts of the selected object, and the
            # center again.  We don't need the max/min this time.
            WVcount=self.BaseGeom.NumVertices()
            WVs=[]

            CX=0.0
            CY=0.0
            CZ=0.0
            for i in range(WVcount):
                WV=self.BaseGeom.WorldVertex(i)
                X=WV.X()
                CX=CX+WV.X()
                Y=WV.Y()
                CY=CY+WV.Y()
                Z=WV.Z()
                CZ=CZ+WV.Z()
                WVs.append( (i,X,Y,Z) )
            # Locate the center
            CX=CX/float(WVcount)
            CY=CY/float(WVcount)
            CZ=CZ/float(WVcount)

            # Now handle each particle at this frame.
            for (PartName,WVindex) in self.PartNameList: # List includes the sample
                Part = scene.Actor(PartName)
                # Get the parms for convenience
                PartXrotParm =   Part.ParameterByCode(poser.kParmCodeXROT)
                PartYrotParm =   Part.ParameterByCode(poser.kParmCodeYROT)
                PartZrotParm =   Part.ParameterByCode(poser.kParmCodeZROT)

                PartXtranParm =  Part.ParameterByCode(poser.kParmCodeXTRAN)
                PartYtranParm =  Part.ParameterByCode(poser.kParmCodeYTRAN)
                PartZtranParm =  Part.ParameterByCode(poser.kParmCodeZTRAN)

                PartAscaleParm = Part.ParameterByCode(poser.kParmCodeASCALE)

                # Set the Y and X rot with desired degree of randomness. (No point
                # in setting all three; gimbal lock and allstufflikethatthere.)
                PartYrotParm.SetValue(PartYrotParm.Value()+360.0 * RandDir * random.uniform(-1.0,1.0))
                PartXrotParm.SetValue(PartXrotParm.Value()+360.0 * RandDir * random.uniform(-1.0,1.0))

                # Set the scale with desired degree of randomness.
                R=self.DefaultScale + random.uniform(-RandSize, RandSize)
                if R < 0.001: R = 0.001
                PartAscaleParm.SetValue(R)

                # Now place the locations, using vectors from center of
                # container to each vert of container as 'sliders'
                # RandPro in this case should be less than 1 but not much
                # less than 1.  When speed is max, RandPro goes from 0.5 to 1;
                # when 0, RandPro is just 1.
                (i,WVx,WVy,WVz)=WVs[WVindex] # Reget the proper WV
                RandPro = 1.0 - Speed * random.uniform(0, 0.5)
                PartXtranParm.SetValue(WVx*RandPro)
                RandPro = 1.0 - Speed * random.uniform(0, 0.5)
                PartYtranParm.SetValue(WVy*RandPro)
                RandPro = 1.0 - Speed * random.uniform(0, 0.5)
                PartZtranParm.SetValue(WVz*RandPro)

    # - - - - - - - - - - - - - - - - - -

    def HandleGo(self):
        self.SetupAll()
        self.Swarm()
        self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents()
        root.lift()
        root.after(250, self.Update)

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

root = Tk()
app = App(root, '')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
